import os
import mmap
import sha
import sys

from resumr.rar import *
from resumr.sync import attempt_sync, SyncError
from resumr.torrent import Torrent

# These numbers are not terribly important, but the speed is better with a
# good guess.
first_volume_guess = 96
subsequent_volume_guess = 100
# This one is vitally important or you'll lose the block that spans volumes
volume_tail = 18

def _get_filesize(fn):
    return os.stat(fn)[6]

def extract(args=None):
    global first_volume_guess, subsequent_volume_guess

    if args is None:
        args = sys.argv[1:]
    torrentname = args[0]
    filenames = args[1:]
    
    tor = Torrent(torrentname)
    rars = find_largest_rarset(filenames)
    print rars.files
    
    # TODO: pick out the largest file if there are multiple
    total_filesize = tor.data['info']['length']
    bs = tor.data['info']['piece length']
    p = tor.data['info']['pieces']
    shas = [p[i:i+20] for i in range(0, len(p), 20)]
    del p
    
    #TODO: warn if the file already exists
    dest = open(tor.data['info']['name'] + '.recovered', 'wb')

    best_bets = rars.best_bets()

    # See if we can come up with a better guess for the header length
    # Commented since it doesn't work right right now
    #(last_index, (last_file, last_file_size)) = best_bets[-1]
    #if is_complete(last_file) and last_file_size < rars.volume_size:
    #    # The last file is probably the true one out of the set... so come
    #    # up with the average header size across the set.
    #    total_set_size = (last_index + 1) * rars.volume_size
    #    extra_bytes = total_set_size - total_filesize
    #    subsequent_volume_guess = extra_bytes // (last_index + 1)
    #    assert subsequent_volume_guess < 1024

    # Data carried over from the previous file
    leftover_junk = ''
    leftover_junk_chunk = None

    dest_chunk = 0
    last_sum_good = True
    prev_index = None

    for (index, (filename, length)) in best_bets:
        print "Loading", filename
        fo = open(filename, 'rb')
        data = mmap.mmap(fo.fileno(), length, access=mmap.ACCESS_READ)

        if index == 0:
            pos = first_volume_guess
            try:
                pos = attempt_sync(data, pos, shas[0], bs)
            except SyncError:
                print "Couldn't sync first volume, %s" % fileinfo[0]
                continue

            print "First volume offset is", pos
            first_volume_guess = pos
        else:
            if last_sum_good and (prev_index + 1 == index):
                # Just account for the estimated junk
                #pos = bs - len(leftover_junk) + subsequent_volume_guess
                #print "Next full chunk should be at", pos
                pos = subsequent_volume_guess
                print "Continuing chunk with junklen =", len(leftover_junk)
            else:
                # Multiply out where we _should_ be... maybe a bad sync or
                # maybe skipping a volume.  `apos` is the byte we estimate
                # this file starts on
                apos = (rars.volume_size - first_volume_guess) + \
                       ((index - 1) * (rars.volume_size -
                                       subsequent_volume_guess))
                dest_chunk = (apos // bs) + 1
                pos = subsequent_volume_guess + (dest_chunk * bs) - apos
                print "Guessing we start at", pos, apos, dest_chunk
                leftover_junk = ''
                dest.seek(dest_chunk * bs, 0)

        prev_index = index
        num_errors = 0

        while pos + bs <= length:
            # Blocksize for this chunk, accounting for leftovers
            rbs = bs - len(leftover_junk)
            csum = sha.new(leftover_junk + data[pos:pos+rbs])
            if csum.digest() == shas[dest_chunk]:
                last_sum_good = True
                print "Good chunk", dest_chunk, "@", pos
                dest.write(leftover_junk + data[pos:pos+rbs])
                leftover_junk = ''
            else:
                last_sum_good = False
                print "Sync error detected at", pos, "for", dest_chunk
                print "  Junk =", len(leftover_junk), rbs, bs
                oldpos = pos
                try:
                    pos = attempt_sync(data, pos, shas[dest_chunk], bs,
                                       leftover_junk)
                    print "  Succeeded, offset %d" % (pos - oldpos)
                    # Need to reset subsequent guess when possible
                    num_errors = 0
                    # Skip the increments at the end of the loop so next
                    # time through we test on the same chunk again
                    continue
                except SyncError:
                    num_errors += 1
                    if num_errors > 2:
                        print "  Sync failed, skipping remaining %d bytes" % \
                              (length - oldpos)
                        break
                    else:
                        print "  Skip to next"
                    # Only reset the junk if we failed the resync with it
                    leftover_junk = ''
                    dest.seek(bs, 1)

            # Regardless of whether the chunk is good, we still increment
            # these values.  This allows skipping a single bad chunk and
            # resyncing with offset 0, or skipping a slightly missized
            # chunk.  Several consecutive bad chunks are caught a few lines
            # above.
            dest_chunk += 1
            pos += rbs
            rbs = bs # reset block size!

        # Check for incomplete chunk, and add to junk
        if pos + bs > length and last_sum_good:
            #TODO: dynamic volume_tail detection by initally skipping the
            # junk chunk, matching the first full chunk, then reconstructing
            # it.
            leftover_junk = data[pos:-volume_tail]
            leftover_junk_chunk = dest_chunk
            print "Leaving junk", len(leftover_junk), "for", dest_chunk

        fo.close()

    # Not really a problem if it's not the last volume... either we complete
    # the file or we replace a missing chunk with a partial chunk.  Note
    # that the last volume includes the tail as well.
    if last_sum_good and leftover_junk:
        dest.write(leftover_junk)

if __name__ == '__main__':
    extract()
